"""
PrismDAG class definition. This class can be used to manipulate Prism projects / tasks
without the CLI

Table of Contents
- Imports
- Functions / utils
- Class definition
"""


###########
# Imports #
###########

# Standard library imports
import argparse
import os
from pathlib import Path
from typing import Any, Dict, List, Optional

# Prism-specific imports
import prism.constants
import prism.cli.base
import prism.exceptions
from prism.infra import executor as prism_executor
from prism.infra import module as prism_module
from prism.infra import compiler as prism_compiler
from prism.infra.project import PrismProject
import prism.mixins.base
import prism.mixins.compile
import prism.mixins.connect
import prism.mixins.run
import prism.mixins.sys_handler
from prism.parsers import ast_parser
import prism.logging


####################
# Class definition #
####################

class PrismDAG(
    prism.mixins.base.BaseMixin,
    prism.mixins.compile.CompileMixin,
    prism.mixins.connect.ConnectMixin,
    prism.mixins.run.RunMixin,
    prism.mixins.sys_handler.SysHandlerMixin
):
    """
    Class to access Prism infrastructure without the CLI
    """

    def __init__(self,
        project_dir: Path,
        log_level: str = 'warn'
    ):
        self.project_dir = project_dir
        self.log_level = log_level

        # Set the project to None.

        # Check if project is valid
        self._is_valid_project(self.project_dir)

        # Modules directory
        self.modules_dir = self.get_modules_dir(self.project_dir)

        # All modules in project
        self.all_modules = self.get_modules(self.modules_dir)

        # Define run context
        self.run_context = prism.constants.CONTEXT.copy()

        # Set up default logger
        args = argparse.Namespace()
        args.log_level = self.log_level
        prism.logging.set_up_logger(args)

    def _is_valid_project(self, user_project_dir: Path) -> bool:
        """
        Determine if `user_project_dir` is a valid project (i.e., that is has a
        `prism_project.py` file and a `modules` folder)

        args:
            user_project_dir: project path
        raises:
            exception if Prism project is not valid
        """
        os.chdir(user_project_dir)
        temp_project_dir = prism.cli.base.get_project_dir()

        # Inputted project directory is not the same as the computed project directory
        if temp_project_dir != user_project_dir:
            msg = f'no project at `{str(user_project_dir)}, closest project found at `{str(temp_project_dir)}`'   # noqa: E501
            raise prism.exceptions.InvalidProjectException(message=msg)

        # Modules folder is not found
        if not Path(user_project_dir / 'modules').is_dir():
            raise prism.exceptions.InvalidProjectException(
                message=f'`modules` directory not found in `{str(user_project_dir)}`'
            )

        return True

    def connect(self,
        connection_type: str
    ):
        """
        Connect task to an adapter
        """
        # Parse arguments
        if connection_type is None:
            raise prism.exceptions.RuntimeException(
                message='connection type cannot be None'
            )
        if not isinstance(connection_type, str):
            raise prism.exceptions.RuntimeException(
                message='connection type must be a string'
            )
        if connection_type not in prism.constants.VALID_ADAPTERS:
            msg = f'invalid connection type `{connection_type}; must be one of {",".join(prism.constants.VALID_ADAPTERS)}'  # noqa: E501
            raise prism.exceptions.RuntimeException(message=msg)

        # Get profiles filepath
        prism_project = PrismProject(
            project_dir=self.project_dir,
            user_context={},
            which="connect",
            filename="prism_project.py"
        )
        profiles_dir = prism_project.project_dir
        profiles_filepath = profiles_dir / 'profile.yml'
        self.create_connection(connection_type, profiles_filepath)

    def compile(self,
        modules: Optional[List[str]] = None,
        all_downstream: bool = True
    ) -> prism_compiler.CompiledDag:
        """
        Compile the Prism project
        """
        if modules is None:
            module_paths = self.all_modules
        else:
            module_paths = [Path(p) for p in modules]
        self.user_arg_modules_list = module_paths

        # Create compiled directory and compile the DAG
        compiled_dir = self.create_compiled_dir(self.project_dir)
        self.compiled_dir = compiled_dir
        return self.compile_dag(
            self.project_dir,
            self.compiled_dir,
            self.all_modules,
            self.user_arg_modules_list,
            all_downstream
        )

    def run(self,
        modules: Optional[List[str]] = None,
        all_upstream: bool = True,
        all_downstream: bool = False,
        full_tb: bool = True,
        user_context: Optional[Dict[str, Any]] = None
    ):
        """
        Run the Prism project
        """
        # Create PrismProject
        if user_context is None:
            user_context = {}
        prism_project = self.create_project(
            project_dir=self.project_dir,
            user_context=user_context,
            which="run",
            filename="prism_project.py",
        )

        # Compile the DAG
        compiled_dag = self.compile(modules, all_downstream)

        # Create DAG executor and Pipeline objects
        threads = prism_project.thread_count
        dag_executor = prism_executor.DagExecutor(
            self.project_dir,
            compiled_dag,
            all_upstream,
            all_downstream,
            threads,
            user_context
        )
        pipeline = self.create_pipeline(
            prism_project, dag_executor, self.run_context
        )

        # Create args and exec
        output = pipeline.exec(full_tb)
        self.run_context = prism_project.cleanup(self.run_context)
        if output.error_event is not None:
            event = output.error_event
            try:
                # Exception is a PrismException
                exception = event.err
            except AttributeError:
                # All other exceptions
                exception = event.value
            raise exception

    def _get_task_cls_from_namespace(self, module_path: Path):
        """
        Get PrismTask associated with `module_path` from Prism namespace

        args:
            module_path: path to module (relative to `modules/` folder)
        returns:
            task class
        """

        # Name of variable containing the PrismTask for `module_path`
        task_var_name = prism_module.get_task_var_name(module_path)
        task_cls = self.run_context.get(task_var_name)

        # If class is None, then raise an error. Otherwise, return the class
        if task_cls is None:
            msg_list = [
                f'no output found for `{str(module_path)}` in the PrismProject namespace',                     # noqa: E501
                f'check that `{str(module_path)}` has a PrismTask and that `{str(module_path)}` has been run'  # noqa: E501
            ]
            raise prism.exceptions.RuntimeException(message='\n'.join(msg_list))
        return task_cls

    def get_task_output(self,
        module_path: Path,
        bool_run: bool = False,
        **kwargs
    ) -> Any:
        """
        Get output of the task in `module`. If bool_run==True, then run the project
        first. Note that if bool_run==False, then only tasks with a target can be
        parsed.

        args:
            module: path to module (relative to `modules/` folder)
            bool_run: boolean indicating whether to run pipeline first; default is False
            **kwargs: keyword arguments for running
        returns:
            task output
        """

        # The user may have run the project in their script. Try retrieving the output
        try:
            task_cls = self._get_task_cls_from_namespace(module_path)
            return task_cls.get_output()

        except prism.exceptions.RuntimeException:
            # If project is run, then any output can be retrieved from any task
            if bool_run:
                self.run(**kwargs)
                task_cls = self._get_task_cls_from_namespace(module_path)
                return task_cls.get_output()

            # If project is not run, then only targets can be retrieved
            else:
                parsed_ast_module = ast_parser.AstParser(module_path, self.modules_dir)
                prism_task_cls = parsed_ast_module.get_prism_task_node(
                    parsed_ast_module.classes, parsed_ast_module.bases
                )
                if prism_task_cls is None:
                    return None
                prism_task_cls_name = prism_task_cls.name

                # We need to update sys.path to include all paths in SYS_PATH_CONF,
                # since some target locations may depend on vars stored in modules
                # imported from directories contained therein.
                user_context = kwargs.get("user_context", {})
                prism_project = self.create_project(
                    project_dir=self.project_dir,
                    user_context=user_context,
                    which="run",
                    filename="prism_project.py"
                )
                self.add_paths_to_sys_path(
                    prism_project.sys_path_config,
                    prism_project.run_context
                )

                # Execute the class definition code
                exec(parsed_ast_module.module_str, prism_project.run_context)
                task = prism_project.run_context[prism_task_cls_name](False)  # type: ignore # noqa: E501

                # No need for an actual task_manager/hooks, since we're only accessing
                # the target
                task.set_hooks(None)
                task.set_task_manager(None)
                task.exec()
                output = task.get_output()
                return output

    def get_pipeline_output(self, bool_run: bool = False) -> Any:
        """
        Get pipeline output, defined as the output associated with the last task in the
        project

        args:
            bool_run: boolean indicating whether to run the pipeline
        returns:
            output associated with last task in the project
        """
        # Compile the project and get the last task
        compiled_dag = self.compile()
        last_task = compiled_dag.topological_sort[-1]
        return self.get_task_output(last_task, bool_run)



"""
Prism Module class

Table of Contents
- Imports
- Class definition
"""

###########
# Imports #
###########

# Standard library imports
from pathlib import Path
from typing import Any, Dict
from types import ModuleType

# Prism-specific imports
import prism.exceptions
from prism.infra.task_manager import PrismTaskManager
from prism.infra.hooks import PrismHooks
from prism.infra.manifest import ModuleManifest
from prism.parsers.ast_parser import AstParser


#####################
# Functions / utils #
#####################

def get_task_var_name(module_path: Path) -> str:
    """
    Retrieve the variable used to store the PrismTask in `module_path` in our namespace

    args:
        module_path: path to module, relative to `modules/`
    returns:
        variable name
    """
    task_var_name = str(module_path).replace('/', '_').replace('.py', '')
    return task_var_name


####################
# Class definition #
####################

class CompiledModule:
    """
    Class for defining and executing a single compiled module
    """

    def __init__(self,
        module_relative_path: Path,
        module_full_path: Path,
        module_manifest: ModuleManifest
    ):
        self.module_relative_path = module_relative_path
        self.module_full_path = module_full_path
        with open(self.module_full_path, 'r') as f:
            self.module_str = f.read()
        f.close()

        # Module as an AST
        parent_path = Path(str(module_full_path).replace(str(module_relative_path), ''))
        self.ast_parser = AstParser(self.module_relative_path, parent_path)

        # Module name
        self.name = str(self.module_relative_path)

        # Set manifest
        self.module_manifest = module_manifest
        self.refs = self._check_manifest(self.module_manifest)

    def _check_manifest(self, module_manifest: ModuleManifest):
        """
        Check manifest and return list of refs associated with compiled
        module
        """
        refs = []
        manifest_refs = module_manifest.manifest_dict["refs"]
        for ref_obj in manifest_refs:
            refs.append(ref_obj["source"])
        if len(refs) == 1:
            refs = refs[0]
        return refs

    def grab_retries_metadata(self):
        """
        Grab retry metadata, including:
            1. How many retries to undertake
            2. The delay between retries
        """
        retries = self.ast_parser.get_variable_assignments(
            self.ast_parser.ast_module, 'RETRIES'
        )
        retry_delay_seconds = self.ast_parser.get_variable_assignments(
            self.ast_parser.ast_module, 'RETRY_DELAY_SECONDS'
        )
        if retries is None:
            retries = 0
        if retry_delay_seconds is None:
            retry_delay_seconds = 0
        return retries, retry_delay_seconds

    def instantiate_module_class(self,
        run_context: Dict[Any, Any],
        task_manager: PrismTaskManager,
        hooks: PrismHooks,
        explicit_run: bool = True,
        user_context: Dict[Any, Any] = {}
    ):
        """
        Instantiate PrismTask child from module

        args:
            run_context: globals dictionary
            task_manager: PrismTaskManager object
            hooks: PrismHooks object
            explicit run: boolean indicating whether to run the Task. Default is True
        returns:
            variable used to store task instantiation
        """
        # Get prism class from module
        prism_task_class = self.ast_parser.get_prism_task_node(
            self.ast_parser.classes, self.ast_parser.bases
        )
        if prism_task_class is None:
            raise prism.exceptions.ParserException(
                message=f"no PrismTask in `{str(self.module_relative_path)}`"
            )
        prism_task_class_name = prism_task_class.name

        # Variable name should just be the name of the module itself. A project
        # shouldn't contain duplicate modules.
        task_var_name = get_task_var_name(self.module_relative_path)

        # If a user context is specified, we need to make sure that the prism_project
        # variables are overridden by whatever the user provides. We need to ensure that
        # these variables are changed GLOBALLY, i.e., in all functions that utilize
        # prism_project variables and all files that import the prism_project. The
        # easiest way to do this is to to import the prism_project.py file before
        # executing the module string and make the necessary adjustments.
        if user_context != {}:

            # By importing the prism_project, we take advantage of Python's import
            # caching. That is, if we execute a module that imports prism_project,
            # Python will see that prism_project has already been imported and will not
            # re-import it and overwrite the user context.
            exec("import prism_project", run_context)

            # Get Prism project and update internal vars
            prism_project_alias = ""
            for k, v in run_context.items():
                if isinstance(v, ModuleType):
                    if v.__name__ == "prism_project":
                        prism_project_alias = k
            if prism_project_alias != "":
                for user_k, user_v in user_context.items():
                    setattr(run_context[prism_project_alias], user_k, user_v)

        # Execute class definition and create task
        exec(self.module_str, run_context)
        run_context[task_var_name] = run_context[prism_task_class_name](explicit_run)

        # Set task manager and hooks
        run_context[task_var_name].set_task_manager(task_manager)
        run_context[task_var_name].set_hooks(hooks)

        # Return name of variable used to store task instantiation
        return task_var_name

    def exec(self,
        run_context: Dict[Any, Any],
        task_manager: PrismTaskManager,
        hooks: PrismHooks,
        explicit_run: bool = True,
        user_context: Dict[Any, Any] = {}
    ) -> PrismTaskManager:
        """
        Execute module
        """
        task_var_name = self.instantiate_module_class(
            run_context, task_manager, hooks, explicit_run, user_context
        )

        # Execute the task
        run_context[task_var_name].exec()
        task_manager.upstream[self.name] = run_context[task_var_name]
        return task_manager



from prism.task import PrismTask
import prism.target as PrismTarget

class Modulee(PrismTask):

    def run(self, tasks, hooks):
        return tasks.ref('moduleA.py') + tasks.ref('moduleC.py') + tasks.ref('moduleD.py') + " This is module E."


# EOF


"""
Prism constants.
"""

###########
# Imports #
###########

import os
from pathlib import Path
import sys
import builtins


#############
# Constants #
#############

# Version number
VERSION = '0.1.7rc3'

# Root directory of project
ROOT_DIR = str(Path(os.path.dirname(__file__)).parent)

# Files to ignore when instantiating Prism project
IGNORE_FILES = ["__pycache__", '*checkpoint.ipynb', '.ipynb_checkpoints']

# Profile/adapter constants
VALID_PROFILE_KEYS = ["adapters"]
VALID_ADAPTERS = [
    "bigquery",
    "dbt",
    "postgres",
    "pyspark",
    "redshift",
    "snowflake",
    "trino",
]

# Context
CONTEXT = {
    '__builtins__': builtins,
    '__name__': '__main__'
}

# Internal names for task_manager and hooks
INTERNAL_TASK_MANAGER_VARNAME = '__PRISM_TASK_MANAGER__'
INTERNAL_HOOKS_VARNAME = '__PRISM_HOOKS__'

# Python version
PYTHON_VERSION = sys.version_info

# Trigger types
VALID_TRIGGER_TYPES = ["function", "prism_project"]



"""
Target class definition.

Table of Contents
- Imports
- Class definitions
- Target decorators
"""

###########
# Imports #
###########

# Prism imports
import prism.exceptions
import prism.infra.hooks
import prism.infra.task_manager


#####################
# Class definitions #
#####################

class PrismTarget:

    def __init__(self, obj, loc, hooks):
        self.obj = obj
        self.loc = loc
        self.hooks = hooks

    def save(self):
        raise prism.exceptions.RuntimeException(message="`save` method not implemented")


class PySparkParquet(PrismTarget):

    def save(self, **kwargs):
        self.obj.write.parquet(self.loc, **kwargs)


class PandasCsv(PrismTarget):

    def save(self, **kwargs):
        self.obj.to_csv(self.loc, **kwargs)


class NumpyTxt(PrismTarget):

    def save(self, **kwargs):
        import numpy as np
        np.savetxt(self.loc, self.obj, **kwargs)


class Txt(PrismTarget):

    def save(self, **kwargs):
        with open(self.loc, "w") as f:
            f.write(self.obj, **kwargs)
        f.close()


class MatplotlibPNG(PrismTarget):

    def save(self, **kwargs):
        self.obj.savefig(self.loc, **kwargs)



"""
BaseEventManager class definition

Table of Contents
- Imports
- Class definition
"""

###########
# Imports #
###########

# Standard library imports
from dataclasses import dataclass
import sys
import time
from typing import Any, Callable, List, Optional, Union

# Prism-specific imports
import prism.logging
import prism.exceptions
from prism.logging import Event, fire_console_event, fire_empty_line_event
from prism.ui import RED, GREEN, BLUE, RESET


####################
# Class definition #
####################

@dataclass
class EventManagerOutput:
    outputs: Any
    event_to_fire: Optional[prism.logging.Event]
    event_list: List[prism.logging.Event]


class BaseEventManager:
    """
    Class for managing logging events fired by tasks and associated
    functions
    """

    def __init__(self,
        idx: Union[None, int],
        total: Union[None, int],
        name: str,
        full_tb: bool,
        func: Callable[..., Any]
    ):
        self.idx = idx
        self.total = total
        self.name = name
        self.full_tb = full_tb
        self.func = func

    def fire_running_exec_event(self,
        event_list: List[prism.logging.Event]
    ):
        """
        Create ExecutionEvent informing user of task execution
        """
        e = prism.logging.ExecutionEvent(
            msg=f"RUNNING {BLUE}{self.name}{RESET}",
            num=self.idx,
            total=self.total,
            status="RUN",
            execution_time=None
        )
        event_list = fire_console_event(e, event_list, log_level='info')
        return event_list

    def fire_success_exec_event(self,
        start_time: float,
        event_list: List[prism.logging.Event]
    ) -> List[Event]:
        """
        Create ExecutionEvent informing user of successful task execution
        """
        execution_time = time.time() - start_time
        e = prism.logging.ExecutionEvent(
            msg=f"{GREEN}FINISHED{RESET} {BLUE}{self.name}{RESET}",
            num=self.idx,
            total=self.total,
            status="DONE",
            execution_time=execution_time
        )
        event_list = fire_console_event(e, event_list, log_level='info')
        return event_list

    def fire_error_exec_event(self,
        start_time: float,
        event_list: List[prism.logging.Event]
    ) -> List[Event]:
        """
        Create ExecutionEvent informing user of error in task execution
        """
        execution_time = time.time() - start_time
        e = prism.logging.ExecutionEvent(
            msg=f"{RED}ERROR{RESET} {BLUE}{self.name}{RESET}",
            num=self.idx,
            total=self.total,
            status="ERROR",
            execution_time=execution_time
        )

        # Set the log-level to `info`. We'll fire the actual error using log-level
        # `error`.
        event_list = fire_console_event(e, event_list, log_level='error')
        return event_list

    def run(self, **kwargs):
        """
        Execute task using inputted function inputted during initialization
        """
        return self.func(**kwargs)

    def manage_events_during_run(self,
        event_list: List[prism.logging.Event],
        fire_exec_events=True,
        fire_empty_line_events=True,
        **kwargs
    ) -> EventManagerOutput:
        """
        Fire relevant event managers
        """

        start_time = time.time()
        if fire_exec_events:
            event_list = self.fire_running_exec_event(event_list)

        # Execute task
        try:
            outputs = self.run(**kwargs)
            if fire_exec_events:
                event_list = self.fire_success_exec_event(start_time, event_list)

            # Return output of task execution
            return EventManagerOutput(outputs, None, event_list)

        # If PrismException, then create PrismExceptionErrorEvent
        except prism.exceptions.PrismException as err:
            if fire_exec_events:
                event_list = self.fire_error_exec_event(start_time, event_list)
            if fire_empty_line_events:
                event_list = fire_empty_line_event(event_list)
            prism_exception_event = prism.logging.PrismExceptionErrorEvent(
                err, self.name
            )
            return EventManagerOutput(0, prism_exception_event, event_list)

        # If SyntaxError, then create ExecutionSyntaxErrorEvent
        except SyntaxError:
            if fire_exec_events:
                event_list = self.fire_error_exec_event(start_time, event_list)
            if fire_empty_line_events:
                event_list = fire_empty_line_event(event_list)
            exc_type, exc_value, exc_tb = sys.exc_info()
            if self.full_tb:
                syntax_error_event = prism.logging.ExecutionSyntaxErrorEvent(
                    self.name, exc_type, exc_value, exc_tb, True
                )
            else:
                syntax_error_event = prism.logging.ExecutionSyntaxErrorEvent(
                    self.name, exc_type, exc_value, exc_tb, False
                )
            return EventManagerOutput(0, syntax_error_event, event_list)

        # If any other Exception, then create ExecutionErrorEvent
        except Exception:
            if fire_exec_events:
                event_list = self.fire_error_exec_event(start_time, event_list)
            if fire_empty_line_events:
                event_list = fire_empty_line_event(event_list)
            exc_type, exc_value, exc_tb = sys.exc_info()
            if self.full_tb:
                exception_event = prism.logging.ExecutionErrorEvent(
                    self.name, exc_type, exc_value, exc_tb, True
                )
            else:
                exception_event = prism.logging.ExecutionErrorEvent(
                    self.name, exc_type, exc_value, exc_tb, False
                )
            return EventManagerOutput(0, exception_event, event_list)



from prism.task import PrismTask
import prism.target as PrismTarget

class Moduled(PrismTask):

    def run(self, tasks, hooks):
        return tasks.ref('moduleB.py') + tasks.ref('moduleA.py') + tasks.ref('moduleC.py') + " This is module D."


# EOF


"""
Unit testing for functions used to parse {{ mod }} references in the modules and
construct the DAG.

Table of Contents:
- Imports
- Compile task instance to call functions
- Test case class definition
"""


###########
# Imports #
###########

# Standard library imports
import unittest
from pathlib import Path
from typing import Union

# Prism imports
import prism.exceptions
from prism.infra import compiler
from prism.tests.unit.test_all_things_dag import TASK_REF_TEST_CASES
from prism.tests.unit.test_all_things_dag.task_ref_3nodes import TASK_REF_3NODES_LIST
from prism.tests.unit.test_all_things_dag.task_ref_5nodes import TASK_REF_5NODES_LIST
from prism.tests.unit.test_all_things_dag.task_ref_15nodes import TASK_REF_15NODES_LIST
from prism.tests.unit.test_all_things_dag.task_ref_norefs import TASK_REF_NOREFS_LIST
from prism.tests.unit.test_all_things_dag.task_ref_selfref import TASK_REF_SELFREF_LIST
from prism.tests.unit.test_all_things_dag.dag_cycle import DAG_CYCLE_LIST

# Make sure that list of modules are actual PosixPath objects
TASK_REF_3NODES_LIST = [Path(p) for p in TASK_REF_3NODES_LIST]
TASK_REF_5NODES_LIST = [Path(p) for p in TASK_REF_5NODES_LIST]
TASK_REF_15NODES_LIST = [Path(p) for p in TASK_REF_15NODES_LIST]
TASK_REF_NOREFS_LIST = [Path(p) for p in TASK_REF_NOREFS_LIST]
TASK_REF_SELFREF_LIST = [Path(p) for p in TASK_REF_SELFREF_LIST]
DAG_CYCLE_LIST = [Path(p) for p in DAG_CYCLE_LIST]


###########################################
# Compile task instance to call functions #
###########################################

# Task
dag_compiler = compiler.DagCompiler(TASK_REF_TEST_CASES, None, None, None, False)

# Constants
TASK_REF_3NODES_DIR = Path(TASK_REF_TEST_CASES) / 'task_ref_3nodes'
TASK_REF_5NODES_DIR = Path(TASK_REF_TEST_CASES) / 'task_ref_5nodes'
TASK_REF_15NODES_DIR = Path(TASK_REF_TEST_CASES) / 'task_ref_15nodes'
TASK_REF_NOREFS_DIR = Path(TASK_REF_TEST_CASES) / 'task_ref_norefs'
TASK_REF_SELFREF_DIR = Path(TASK_REF_TEST_CASES) / 'task_ref_selfref'
DAG_CYCLE_DIR = Path(TASK_REF_TEST_CASES) / 'dag_cycle'


##############################
# Test case class definition #
##############################

class TestAllDagFunctions(unittest.TestCase):

    def _convert_task_refs_to_str(self, task_refs: dict):
        task_refs_str = {}
        for k, v in task_refs.items():
            new_k = str(k)
            if isinstance(v, Path):
                new_v = str(v)
            elif isinstance(v, list):
                new_v = [str(p) for p in v]
            elif v is None:
                new_v = None
            else:
                raise Exception
            task_refs_str[new_k] = new_v
        return task_refs_str

    def _convert_nodes_to_str(self, nodes: list):
        return [str(p) for p in nodes]

    def _convert_edges_to_str(self, edges: list):
        edges_str = []
        for e in edges:
            new_edge = (str(e[0]), str(e[1]))
            edges_str.append(new_edge)
        return edges_str

    def _convert_topsort_to_str(self, topsort: list):
        return [str(p) for p in topsort]

    ###################
    # Mod ref parsing #
    ###################

    def test_task_ref_varying_complexity(self):
        """
        Parse mod refs with different node counts and dependency complexities. These
        should not result in an error.
        """
        task_ref_3nodes_modrefs = dag_compiler.parse_task_refs(
            TASK_REF_3NODES_LIST, TASK_REF_3NODES_DIR
        )
        task_ref_5nodes_modrefs = dag_compiler.parse_task_refs(
            TASK_REF_5NODES_LIST, TASK_REF_5NODES_DIR
        )
        task_ref_15nodes_modrefs = dag_compiler.parse_task_refs(
            TASK_REF_15NODES_LIST, TASK_REF_15NODES_DIR
        )

        # Expected mod refs. The order here is based on the order of the
        # `TASK_REF_XNODES_LIST` and the order of {{ mod }} within the modules
        # themselves.
        expected_3nodes_modrefs = {
            Path('module01.py'): None,
            Path('module02.py'): Path('module01.py'),
            Path('module03.py'): Path('module02.py')
        }

        expected_5nodes_modrefs = {
            Path('moduleA.py'): None,
            Path('moduleB.py'): Path('moduleA.py'),
            Path('moduleC.py'): Path('moduleA.py'),
            Path('moduleD.py'): [
                Path('moduleB.py'),
                Path('moduleA.py'),
                Path('moduleC.py')
            ],
            Path('moduleE.py'): [
                Path('moduleA.py'),
                Path('moduleC.py'),
                Path('moduleD.py')
            ],
        }

        expected_15nodes_modrefs = {
            Path('module01.py'): None,
            Path('module02.py'): Path('module01.py'),
            Path('module03.py'): Path('module01.py'),
            Path('module04.py'): [Path('module02.py'), Path('module03.py')],
            Path('module05.py'): Path('module01.py'),
            Path('module06.py'): Path('module05.py'),
            Path('module07.py'): [Path('module04.py'), Path('module06.py')],
            Path('module08.py'): Path('module01.py'),
            Path('module09.py'): [Path('module05.py'), Path('module08.py')],
            Path('module10.py'): Path('module01.py'),
            Path('module11.py'): Path('module10.py'),
            Path('module12.py'): Path('module10.py'),
            Path('module13.py'): Path('module10.py'),
            Path('module14.py'): Path('module11.py'),
            Path('module15.py'): Path('module11.py')
        }

        # Sort mod ref values to test equality
        def _sort_values(task_ref_dict: dict):
            return {k: sorted(v) if isinstance(v, list) else v for k, v in task_ref_dict.items()}  # noqa: E501

        expected_3nodes_modrefs_sorted = _sort_values(expected_3nodes_modrefs)
        expected_5nodes_modrefs_sorted = _sort_values(expected_5nodes_modrefs)
        expected_15nodes_modrefs_sorted = _sort_values(expected_15nodes_modrefs)
        task_ref_3nodes_modrefs_sorted = _sort_values(task_ref_3nodes_modrefs)
        task_ref_5nodes_modrefs_sorted = _sort_values(task_ref_5nodes_modrefs)
        task_ref_15nodes_modrefs_sorted = _sort_values(task_ref_15nodes_modrefs)

        # Asserts
        self.assertEqual(
            expected_3nodes_modrefs_sorted, task_ref_3nodes_modrefs_sorted
        )
        self.assertEqual(
            expected_5nodes_modrefs_sorted, task_ref_5nodes_modrefs_sorted
        )
        self.assertEqual(
            expected_15nodes_modrefs_sorted, task_ref_15nodes_modrefs_sorted
        )

    def test_no_task_refs(self):
        """
        Modules without any mod refs will not throw an error
        """
        task_ref_nomod_modrefs = dag_compiler.parse_task_refs(
            TASK_REF_NOREFS_LIST, TASK_REF_NOREFS_DIR
        )
        expected_nomod_modrefs = {
            Path('moduleA.py'): None,
            Path('moduleB.py'): None,
            Path('moduleC.py'): None,
            Path('moduleD.py'): None,
            Path('moduleE.py'): None
        }

        # Asserts
        self.assertEqual(expected_nomod_modrefs, task_ref_nomod_modrefs)

    def test_self_ref(self):
        """
        Self-references with mod throw a ParserException
        """
        with self.assertRaises(prism.exceptions.ParserException) as cm:
            dag_compiler.parse_task_refs(TASK_REF_SELFREF_LIST, TASK_REF_SELFREF_DIR)
        expected_msg = "self-references found in `moduleB.py`"
        self.assertEqual(expected_msg, str(cm.exception))

    #######################
    # Create nodes, edges #
    #######################

    def test_create_nodes_edges_varying_complexity(self):
        """
        Use parsed mod refs (which we know from above should work as expected) to create
        nodes and edges for 3-, 5-, and 15-node graphs. These should not throw any
        errors.
        """
        task_ref_3nodes_modrefs = dag_compiler.parse_task_refs(
            TASK_REF_3NODES_LIST, TASK_REF_3NODES_DIR
        )
        task_ref_5nodes_modrefs = dag_compiler.parse_task_refs(
            TASK_REF_5NODES_LIST, TASK_REF_5NODES_DIR
        )
        task_ref_15nodes_modrefs = dag_compiler.parse_task_refs(
            TASK_REF_15NODES_LIST, TASK_REF_15NODES_DIR
        )

        # Create nodes and edges
        nodes_3nodes, edges_3nodes = dag_compiler.create_nodes_edges(
            task_ref_3nodes_modrefs
        )
        nodes_5nodes, edges_5nodes = dag_compiler.create_nodes_edges(
            task_ref_5nodes_modrefs
        )
        nodes_15nodes, edges_15nodes = dag_compiler.create_nodes_edges(
            task_ref_15nodes_modrefs
        )

        # Nodes should just be the full list of modules
        self.assertEqual(TASK_REF_3NODES_LIST, nodes_3nodes)
        self.assertEqual(TASK_REF_5NODES_LIST, nodes_5nodes)
        self.assertEqual(TASK_REF_15NODES_LIST, nodes_15nodes)

        # Expected edges
        expected_edges_3nodes = [
            (Path('module01.py'), Path('module02.py')),
            (Path('module02.py'), Path('module03.py'))
        ]

        expected_edges_5nodes = [
            (Path('moduleA.py'), Path('moduleB.py')),
            (Path('moduleA.py'), Path('moduleC.py')),
            (Path('moduleA.py'), Path('moduleD.py')),
            (Path('moduleA.py'), Path('moduleE.py')),
            (Path('moduleB.py'), Path('moduleD.py')),
            (Path('moduleC.py'), Path('moduleD.py')),
            (Path('moduleC.py'), Path('moduleE.py')),
            (Path('moduleD.py'), Path('moduleE.py'))
        ]

        expected_edges_15nodes = [
            (Path('module01.py'), Path('module10.py')),
            (Path('module01.py'), Path('module02.py')),
            (Path('module01.py'), Path('module03.py')),
            (Path('module01.py'), Path('module05.py')),
            (Path('module01.py'), Path('module08.py')),
            (Path('module10.py'), Path('module11.py')),
            (Path('module10.py'), Path('module12.py')),
            (Path('module10.py'), Path('module13.py')),
            (Path('module02.py'), Path('module04.py')),
            (Path('module03.py'), Path('module04.py')),
            (Path('module05.py'), Path('module06.py')),
            (Path('module05.py'), Path('module09.py')),
            (Path('module08.py'), Path('module09.py')),
            (Path('module11.py'), Path('module14.py')),
            (Path('module11.py'), Path('module15.py')),
            (Path('module04.py'), Path('module07.py')),
            (Path('module06.py'), Path('module07.py'))
        ]

        # Asserts
        self.assertEqual(set(expected_edges_3nodes), set(edges_3nodes))
        self.assertEqual(set(expected_edges_5nodes), set(edges_5nodes))
        self.assertEqual(set(expected_edges_15nodes), set(edges_15nodes))

    def test_create_nodes_edges_no_modrefs(self):
        """
        Nodes/edges are created as expected even in situations where scripts have no mod
        refs
        """
        task_ref_nomod_modrefs = dag_compiler.parse_task_refs(
            TASK_REF_NOREFS_LIST, TASK_REF_NOREFS_DIR
        )
        nodes_nomod, edges_nomod = dag_compiler.create_nodes_edges(
            task_ref_nomod_modrefs
        )

        # Node should just be the full list of nodes
        self.assertEqual(TASK_REF_NOREFS_LIST, nodes_nomod)

        # Edge list should be empty
        self.assertEqual([], edges_nomod)

    ##############
    # Create DAG #
    ##############

    def test_create_dag_varying_complexity(self):
        """
        Create topsort for DAGs with different node counts and dependency complexities.
        These should not result in an error.
        """
        # Parse mod refs
        task_ref_3nodes_modrefs = dag_compiler.parse_task_refs(
            TASK_REF_3NODES_LIST, TASK_REF_3NODES_DIR
        )
        task_ref_5nodes_modrefs = dag_compiler.parse_task_refs(
            TASK_REF_5NODES_LIST, TASK_REF_5NODES_DIR
        )
        task_ref_15nodes_modrefs = dag_compiler.parse_task_refs(
            TASK_REF_15NODES_LIST, TASK_REF_15NODES_DIR
        )

        # Create nodes and edges
        nodes_3nodes, edges_3nodes = dag_compiler.create_nodes_edges(
            task_ref_3nodes_modrefs
        )
        nodes_5nodes, edges_5nodes = dag_compiler.create_nodes_edges(
            task_ref_5nodes_modrefs
        )
        nodes_15nodes, edges_15nodes = dag_compiler.create_nodes_edges(
            task_ref_15nodes_modrefs
        )

        # Create DAG
        dag_compiler.create_dag(nodes_3nodes, edges_3nodes)
        dag_compiler.create_dag(nodes_5nodes, edges_5nodes)
        dag_compiler.create_dag(nodes_15nodes, edges_15nodes)

    def test_dag_no_modrefs(self):
        """
        DAG is created as expected even in situations where scripts have no mod refs
        """
        task_ref_nomod_modrefs = dag_compiler.parse_task_refs(
            TASK_REF_NOREFS_LIST, TASK_REF_NOREFS_DIR
        )
        nodes_nomod, edges_nomod = dag_compiler.create_nodes_edges(
            task_ref_nomod_modrefs
        )
        dag_compiler.create_dag(nodes_nomod, edges_nomod)

    def test_create_dag_cycle(self):
        """
        Cycle in DAG throws an error
        """
        # Parse mod refs
        dag_cycle_modrefs = dag_compiler.parse_task_refs(DAG_CYCLE_LIST, DAG_CYCLE_DIR)
        expected_dag_cycle_modrefs = {
            Path('moduleA.py'): None,
            Path('moduleB.py'): [Path('moduleA.py'), Path('moduleE.py')],
            Path('moduleC.py'): Path('moduleA.py'),
            Path('moduleD.py'): [
                Path('moduleB.py'),
                Path('moduleA.py'),
                Path('moduleC.py')
            ],
            Path('moduleE.py'): [
                Path('moduleA.py'),
                Path('moduleC.py'),
                Path('moduleD.py')
            ]
        }

        # Sort values of dict for testing equality
        expected_dag_cycle_modrefs_sorted = {k: sorted(v) if isinstance(v, list) else v for k, v in expected_dag_cycle_modrefs.items()}  # noqa: E501
        dag_cycle_modrefs_sorted = {k: sorted(v) if isinstance(v, list) else v for k, v in dag_cycle_modrefs.items()}  # noqa: E501
        self.assertEqual(expected_dag_cycle_modrefs_sorted, dag_cycle_modrefs_sorted)

        # Create nodes and edges
        nodes_dag_cycle, edges_dag_cycle = dag_compiler.create_nodes_edges(
            dag_cycle_modrefs
        )
        expected_edges_dag_cycle = [
            (Path('moduleA.py'), Path('moduleB.py')),
            (Path('moduleA.py'), Path('moduleC.py')),
            (Path('moduleA.py'), Path('moduleD.py')),
            (Path('moduleA.py'), Path('moduleE.py')),
            (Path('moduleB.py'), Path('moduleD.py')),
            (Path('moduleC.py'), Path('moduleD.py')),
            (Path('moduleC.py'), Path('moduleE.py')),
            (Path('moduleD.py'), Path('moduleE.py')),
            (Path('moduleE.py'), Path('moduleB.py'))
        ]
        self.assertEqual(set(DAG_CYCLE_LIST), set(nodes_dag_cycle))
        self.assertEqual(set(expected_edges_dag_cycle), set(edges_dag_cycle))

        # Create DAG
        with self.assertRaises(prism.exceptions.DAGException) as cm:
            dag_compiler.create_dag(nodes_dag_cycle, edges_dag_cycle)
        self.assertTrue("invalid DAG, cycle found in " in str(cm.exception))
        self.assertTrue("moduleB.py" in str(cm.exception))
        self.assertTrue("moduleE.py" in str(cm.exception))

    #########################
    # Get node dependencies #
    #########################

    def test_node_dependencies_3nodes(self):
        """
        Get node dependencies from DAGs with 3 nodes.
        """
        task_ref_3nodes_modrefs = dag_compiler.parse_task_refs(
            TASK_REF_3NODES_LIST, TASK_REF_3NODES_DIR
        )
        nodes_3nodes, edges_3nodes = dag_compiler.create_nodes_edges(
            task_ref_3nodes_modrefs
        )
        dag_3nodes = dag_compiler.create_dag(nodes_3nodes, edges_3nodes)

        # Node dependencies
        node_deps_1 = dag_compiler.get_node_dependencies(
            dag_3nodes, [Path('module01.py')]
        )
        node_deps_2 = dag_compiler.get_node_dependencies(
            dag_3nodes, [Path('module02.py')]
        )
        node_deps_3 = dag_compiler.get_node_dependencies(
            dag_3nodes, [Path('module03.py')]
        )
        node_deps_1_2 = dag_compiler.get_node_dependencies(
            dag_3nodes, [Path('module01.py'), Path('module02.py')]
        )

        # Expected node dependencies
        expected_node_deps_1 = [Path('module01.py')]
        expected_node_deps_2 = [Path('module01.py'), Path('module02.py')]
        expected_node_deps_3 = [Path('module01.py'), Path('module02.py'), Path('module03.py')]  # noqa: E501
        expected_node_deps_1_2 = [Path('module01.py'), Path('module02.py')]

        # Asserts
        self.assertEqual(set(expected_node_deps_1), set(node_deps_1))
        self.assertEqual(set(expected_node_deps_2), set(node_deps_2))
        self.assertEqual(set(expected_node_deps_3), set(node_deps_3))
        self.assertEqual(set(expected_node_deps_1_2), set(node_deps_1_2))

    def test_node_dependencies_5nodes(self):
        """
        Get node dependencies from DAGs with 5 nodes.
        """
        task_ref_5nodes_modrefs = dag_compiler.parse_task_refs(
            TASK_REF_5NODES_LIST, TASK_REF_5NODES_DIR
        )
        nodes_5nodes, edges_5nodes = dag_compiler.create_nodes_edges(
            task_ref_5nodes_modrefs
        )
        dag_5nodes = dag_compiler.create_dag(nodes_5nodes, edges_5nodes)

        # Node dependencies
        node_deps_e = dag_compiler.get_node_dependencies(
            dag_5nodes, [Path('moduleE.py')]
        )
        expected_node_deps_e = [
            Path('moduleA.py'),
            Path('moduleB.py'),
            Path('moduleC.py'),
            Path('moduleD.py'),
            Path('moduleE.py'),
        ]
        self.assertEqual(set(expected_node_deps_e), set(node_deps_e))

    def test_node_dependencies_15nodes(self):
        """
        Get node dependencies from DAGs with 15 nodes.
        """
        task_ref_15nodes_modrefs = dag_compiler.parse_task_refs(
            TASK_REF_15NODES_LIST, TASK_REF_15NODES_DIR
        )
        nodes_15nodes, edges_15nodes = dag_compiler.create_nodes_edges(
            task_ref_15nodes_modrefs
        )
        dag_15nodes = dag_compiler.create_dag(nodes_15nodes, edges_15nodes)

        # Node dependencies
        node_deps_7 = dag_compiler.get_node_dependencies(
            dag_15nodes, [Path('module07.py')]
        )
        expected_node_deps_7 = [
            Path('module07.py'),
            Path('module04.py'),
            Path('module06.py'),
            Path('module02.py'),
            Path('module03.py'),
            Path('module05.py'),
            Path('module01.py')
        ]
        self.assertEqual(set(expected_node_deps_7), set(node_deps_7))

        node_deps_all = dag_compiler.get_node_dependencies(
            dag_15nodes,
            [
                Path('module07.py'),
                Path('module14.py'),
                Path('module15.py'),
                Path('module12.py'),
                Path('module13.py'),
                Path('module09.py')
            ]
        )
        expected_node_deps_all = [
            Path('module01.py'),
            Path('module02.py'),
            Path('module03.py'),
            Path('module04.py'),
            Path('module05.py'),
            Path('module06.py'),
            Path('module07.py'),
            Path('module08.py'),
            Path('module09.py'),
            Path('module10.py'),
            Path('module11.py'),
            Path('module12.py'),
            Path('module13.py'),
            Path('module14.py'),
            Path('module15.py')
        ]
        self.assertEqual(set(expected_node_deps_all), set(node_deps_all))

    ##################
    # Create topsort #
    ##################

    def test_topsort_varying_complexity(self):
        """
        Create topsort for DAGs with different node counts and dependency complexities.
        These should not result in an error.
        """
        _, dag_topsort_3nodes = dag_compiler.create_topsort(
            TASK_REF_3NODES_LIST, TASK_REF_3NODES_LIST, TASK_REF_3NODES_DIR
        )
        _, dag_topsort_5nodes = dag_compiler.create_topsort(
            TASK_REF_5NODES_LIST, TASK_REF_5NODES_LIST, TASK_REF_5NODES_DIR
        )
        _, dag_topsort_15nodes = dag_compiler.create_topsort(
            TASK_REF_15NODES_LIST, TASK_REF_15NODES_LIST, TASK_REF_15NODES_DIR
        )

        # For our simple, 3-node DAG, we know what the topsort list should look like.
        # For the others, there are multiple possible orderings. We initially tried to
        # sort the result of the NetworkX `all_topological_sorts` function returns a
        # generator, but this turned out to be too computationally expensive when the
        # number of possible orderings was high. Therefore, rather than test the value
        # of the list itself, we will confirm that parent nodes appear before their
        # children in the list.
        expected_dag_topsort_3nodes = [
            Path('module01.py'),
            Path('module02.py'),
            Path('module03.py')
        ]
        self.assertEqual(expected_dag_topsort_3nodes, dag_topsort_3nodes)

        # Helper function for testing
        def parent_before_child(parent_list, parent: str, child: Union[str, list]):
            idx1 = parent_list.index(parent)
            if isinstance(child, str):
                idx2 = parent_list.index(child)
                self.assertTrue(idx1 < idx2)
            elif isinstance(child, list):
                for node in child:
                    idx2 = parent_list.index(node)
                    self.assertTrue(idx1 < idx2)

        # 5-node DAG
        parent_before_child(
            dag_topsort_5nodes,
            Path('moduleA.py'),
            [
                Path('moduleB.py'),
                Path('moduleC.py'),
                Path('moduleD.py'),
                Path('moduleE.py')
            ],
        )
        parent_before_child(
            dag_topsort_5nodes, Path('moduleB.py'), Path('moduleD.py'),
        )
        parent_before_child(
            dag_topsort_5nodes,
            Path('moduleC.py'),
            [Path('moduleD.py'), Path('moduleE.py')],
        )
        parent_before_child(
            dag_topsort_5nodes, Path('moduleD.py'), Path('moduleE.py')
        )

        # 15-node DAG
        parent_before_child(
            dag_topsort_15nodes,
            Path('module01.py'),
            [
                Path('module10.py'),
                Path('module02.py'),
                Path('module03.py'),
                Path('module05.py'),
                Path('module08.py')
            ],
        )
        parent_before_child(
            dag_topsort_15nodes,
            Path('module10.py'),
            [Path('module11.py'), Path('module12.py'), Path('module13.py')],
        )
        parent_before_child(
            dag_topsort_15nodes, Path('module02.py'), Path('module04.py')
        )
        parent_before_child(
            dag_topsort_15nodes, Path('module03.py'), Path('module04.py')
        )
        parent_before_child(
            dag_topsort_15nodes,
            Path('module05.py'), [Path('module06.py'), Path('module09.py')]
        )
        parent_before_child(
            dag_topsort_15nodes, Path('module08.py'), Path('module09.py')
        )
        parent_before_child(
            dag_topsort_15nodes,
            Path('module11.py'), [Path('module14.py'), Path('module15.py')]
        )
        parent_before_child(
            dag_topsort_15nodes, Path('module04.py'), Path('module07.py')
        )
        parent_before_child(
            dag_topsort_15nodes, Path('module06.py'), Path('module07.py')
        )

    def test_topsort_no_modrefs(self):
        """
        DAGs with no dependencies should include all modules (assuming a subset of
        modules is not explicitly specified).
        """
        _, dag_topsort_no_modrefs = dag_compiler.create_topsort(
            TASK_REF_NOREFS_LIST, TASK_REF_NOREFS_LIST, TASK_REF_NOREFS_DIR
        )

        # All modules must be in DAG
        for mod in TASK_REF_NOREFS_LIST:
            self.assertTrue(mod in dag_topsort_no_modrefs)

    def test_topsort_subset(self):
        """
        DAGs created with only a subset of nodes are created normally / without error
        """
        _, dag_topsort_15nodes_7 = dag_compiler.create_topsort(
            TASK_REF_15NODES_LIST, [Path('module07.py')], TASK_REF_15NODES_DIR
        )
        _, dag_topsort_15nodes_all = dag_compiler.create_topsort(
            TASK_REF_15NODES_LIST,
            [
                Path('module07.py'),
                Path('module14.py'),
                Path('module15.py'),
                Path('module12.py'),
                Path('module13.py'),
                Path('module09.py')
            ],
            TASK_REF_15NODES_DIR
        )
        _, dag_topsort_no_modrefs = dag_compiler.create_topsort(
            TASK_REF_NOREFS_LIST,
            [Path('moduleB.py'), Path('moduleC.py')],
            TASK_REF_NOREFS_DIR
        )

        # In the DAG with 15 nodes, specify which nodes should be included in the DAG
        # based on the modules for compilation
        actual_topsort_15nodes_7 = [
            Path('module01.py'),
            Path('module02.py'),
            Path('module03.py'),
            Path('module04.py'),
            Path('module05.py'),
            Path('module06.py'),
            Path('module07.py')
        ]
        actual_topsort_15nodes_all = TASK_REF_15NODES_LIST
        self.assertEqual(set(actual_topsort_15nodes_7), set(dag_topsort_15nodes_7))
        self.assertEqual(set(actual_topsort_15nodes_all), set(dag_topsort_15nodes_all))

        # In the DAG without mod refs, there aren't any edges. The only nodes in the DAG
        # will be moduleB and moduleC
        actual_topsort_no_modrefs = [Path('moduleB.py'), Path('moduleC.py')]
        self.assertEqual(set(actual_topsort_no_modrefs), set(dag_topsort_no_modrefs))



from prism.task import PrismTask
import prism.target as PrismTarget

class Module06(prism.task.PrismTask):

    def run(self, tasks, hooks):
        return tasks.ref('module05.py') + "This is module 06. "


# EOF


"""
Mixin classes for SysHandler

Table of Contents
- Imports
- Class definition
"""


###########
# Imports #
###########

# Standard library imports
from typing import Any, Dict, List
from pathlib import Path


####################
# Class definition #
####################


class SysHandlerMixin:
    """
    Class for managing sys.path and sys.modules
    """

    def add_paths_to_sys_path(self,
        paths: List[Path],
        globals_dict: Dict[Any, Any]
    ):
        """
        Add paths in `paths` to sys.path

        args:
            paths: paths to add
            globals_dict: sys namespace
        returns:
            none, this function updates the sys.path in-place
        """
        if 'sys' not in globals_dict.keys():
            exec('import sys', globals_dict)

        # Add the paths before the standard sys.path locations, in case there are any
        # overwrites
        paths_to_add = []
        for p in paths:
            if str(p) not in globals_dict['sys'].path:
                paths_to_add.append(str(p))
        globals_dict['sys'].path = paths_to_add + globals_dict['sys'].path
        return globals_dict

    def remove_paths_from_sys_path(self,
        base_sys_path: List[str],
        paths: List[Path],
        globals_dict: Dict[Any, Any]
    ):
        """
        Remove paths in `paths` to sys.path if they were not already in `base_sys_path`

        args:
            base_sys_path: base sys.path
            paths: paths to remove
            globals_dict: sys namespace
        returns:
            none, this function updates the sys.path in-place
        """
        if 'sys' not in globals_dict.keys():
            exec('import sys', globals_dict)

        for p in paths:
            try:
                if str(p) not in base_sys_path:
                    globals_dict['sys'].path.remove(str(p))
            except ValueError:
                continue
        return globals_dict

    def add_sys_path(self, project_dir: Path, globals_dict: Dict[Any, Any]):
        """
        Add project directory to sys.path
        """
        exec('import sys', globals_dict)
        if str(project_dir) not in globals_dict['sys'].path:
            globals_dict['sys'].path.insert(0, str(project_dir))
        return globals_dict

    def remove_sys_path(self, project_dir: Path, globals_dict: Dict[Any, Any]):
        """
        Remove project directory from sys.path
        """
        globals_dict['sys'].path.remove(str(project_dir))
        return globals_dict

    def remove_project_modules(self,
        base_sys_modules: Dict[str, Any],
        paths: List[Path],
        globals_dict: Dict[Any, Any]
    ):
        """
        Remove paths and dependent modules in `paths` from the sys.paths and
        sys.modules. Make sure to only remove them if they were not part of the base
        configuration. This usually isn't necessary, because prism projects run in their
        own Python session. However, there may be cases where the user runs multiple
        prism projects during the same session (e.g., during integration tests).

        This is definitely not best practice; need to find a better way of doing this.

        args:
            base_sys_modules: base sys.modules
            paths: custom paths (`SYS_PATH_CONF` in prism_project.py)
            globals_dict: globals dictionary
        returns:
            None
        """
        if 'sys' not in globals_dict.keys():
            return globals_dict

        # Iterate through all modules. Only delete modules that (1) originate from a
        # path in `paths`, and (2) did not exist in the `base_sys_modules``
        mods_to_del = []
        for mod_name, mod_obj in globals_dict['sys'].modules.items():
            try:
                if mod_obj.__file__ is None:
                    continue
                elif any([p in paths for p in Path(mod_obj.__file__).parents]):
                    mods_to_del.append(mod_name)
            except AttributeError:
                continue

        # Delete modules
        for mod in mods_to_del:
            del globals_dict['sys'].modules[mod]
        return globals_dict



"""
Postgres adapter class definition

Table of Contents
- Imports
- Class definition
"""

###########
# Imports #
###########

# Standard library imports
import pandas as pd
from typing import Any, Dict
import psycopg2

# Prism-specific imports
from .adapter import Adapter
import prism.exceptions


####################
# Class definition #
####################

class Postgres(Adapter):

    def is_valid_config(self,
        config_dict: Dict[str, str],
        adapter_name: str,
        profile_name: str,
    ) -> bool:
        """
        Check that config dictionary is profile.yml is valid

        args:
            config_dict: config dictionary under PostgresQL adapter in profile.yml
            adapter_name: name assigned to adapter
            profile_name: profile name containing adapter
        returns:
            boolean indicating whether config dictionary in profile.yml is valid
        """

        # Required config vars
        required_config_vars = [
            'type',
            'user',
            'password',
            'port',
            'host',
            'database'
        ]

        # Optional config vars
        optional_config_vars = [
            'autocommit'
        ]

        # Raise an error if:
        #   1. Config doesn't contain any of the required vars or contains additional
        #      config vars
        #   2. Any of the config values are None
        actual_config_vars = []
        for k, v in config_dict.items():
            if k not in required_config_vars and k not in optional_config_vars:
                raise prism.exceptions.InvalidProfileException(
                    message=f'invalid var `{k}` - see `{adapter_name}` adapter in `{profile_name}` profile in profile.yml'  # noqa: E501
                )
            if k in required_config_vars:
                actual_config_vars.append(k)
            if v is None:
                raise prism.exceptions.InvalidProfileException(
                    message=f'var `{k}` cannot be None - see `{adapter_name}` adapter in `{profile_name}` profile in profile.yml'  # noqa: E501
                )
        vars_not_defined = list(set(required_config_vars) - set(actual_config_vars))
        if len(vars_not_defined) > 0:
            v = vars_not_defined.pop()
            raise prism.exceptions.InvalidProfileException(
                message=f'var `{v}` must be defined - see `{adapter_name}` adapter in `{profile_name}` profile in profile.yml'  # noqa: E501
            )

        # If no exception has been raised, return True
        return True

    def create_engine(self,
        adapter_dict: Dict[str, Any],
        adapter_name: str,
        profile_name: str
    ):
        """
        Parse PostgresQL adapter, represented as a dict and return the PostgresQL
        connector object

        args:
            adapter_dict: PostgresQL adapter represented as a dictionary
            adapter_name: name assigned to adapter
            profile_name: profile name containing adapter
        returns:
            PostgresQL connector object
        """

        # Get configuration and check if config is valid
        self.is_valid_config(adapter_dict, adapter_name, profile_name)

        # Create psycopg2 connection
        conn = psycopg2.connect(
            dbname=adapter_dict['database'],
            host=adapter_dict['host'],
            port=adapter_dict['port'],
            user=adapter_dict['user'],
            password=adapter_dict['password']
        )

        # Autocommit. If no autocommit is specified, then set to True
        try:
            autocommit_config = bool(adapter_dict['autocommit'])
            conn.set_session(autocommit=autocommit_config)
        except KeyError:
            conn.set_session(autocommit=True)
        return conn

    def execute_sql(self, query: str, return_type: str) -> pd.DataFrame:
        """
        Execute the SQL query
        """
        # Create cursor for every SQL query -- this ensures thread safety
        cursor = self.engine.cursor()
        cursor.execute(query)
        if return_type == "pandas":
            data = cursor.fetchall()
            cols = []
            for elts in cursor.description:
                cols.append(elts[0])
            df: pd.DataFrame = pd.DataFrame(data=data, columns=cols)
            cursor.close()
            return df
        else:
            cursor.close()



from prism.task import PrismTask

class FirstPrismTask(PrismTask):
    
    def run(self, tasks, hooks):
        return 'hi'


class SecondPrismTask(PrismTask):
    
    def run(self, tasks, hooks):
        return 'hi'


# EOF


"""
PrismTaskManager class

Table of Contents
- Imports
- Class definition
"""

###########
# Imports #
###########

# Standard library imports
from typing import Any, Dict


####################
# Class definition #
####################

class PrismTaskManager:
    """
    PrismTaskManager class. This class manages all tasks in the DAG (and their
    associated outputs). An instance of the task manager is passed to all `run`
    functions (via the kw `tasks`), and users can reference the output of other tasks
    via `tasks.ref('...')`.
    """

    def __init__(self, upstream: Dict[str, Any]):
        self.upstream = upstream

    def ref(self, module: str):
        return self.upstream[module].get_output()



###########
# Imports #
###########

# Prism infrastructure imports
import prism.task
import prism.target
import prism.decorators

# Prism project imports
import prism_project

# Other imports
import time
import pandas as pd


######################
## Class definition ##
######################

class SnowflakeTask(prism.task.PrismTask):

    ## Run
    @prism.decorators.target(type=prism.target.PandasCsv, loc=prism_project.OUTPUT / 'machinery_sample.csv', index=False)
    @prism.decorators.target(type=prism.target.PandasCsv, loc=prism_project.OUTPUT / 'household_sample.csv', index=False)
    def run(self, tasks, hooks):
        """
        Execute task.

        args:
            tasks: used to reference output of other tasks --> tasks.ref('...')
            hooks: built-in Prism hooks. These include:
                - hooks.dbt_ref --> for getting dbt models as a pandas DataFrame
                - hooks.sql     --> for executing sql query using an adapter in profile.yml
                - hooks.spark   --> for accessing SparkSession (if pyspark specified in profile.yml)
        returns:
            task output
        """
        machinery_sql = f"""
        SELECT 
            * 
        FROM "SNOWFLAKE_SAMPLE_DATA"."TPCH_SF1"."CUSTOMER" 
        WHERE 
            C_MKTSEGMENT = 'MACHINERY' 
        LIMIT 50
        """
        machinery_df = hooks.sql(adapter_name="snowflake_base", query=machinery_sql)

        household_sql = f"""
        SELECT 
            * 
        FROM "SNOWFLAKE_SAMPLE_DATA"."TPCH_SF1"."CUSTOMER" 
        WHERE 
            C_MKTSEGMENT = 'HOUSEHOLD' 
        LIMIT 50
        """
        household_df = hooks.sql(adapter_name="snowflake_base", query=household_sql)

        return machinery_df, household_df


# EOF



"""
Prism Exceptions
"""


class PrismException(Exception):
    """
    Parent class for all prismexceptions
    """
    pass


class InvalidProjectException(PrismException):
    """
    Exception raised if project is not valid
    """

    def __init__(self, message):
        self.message = message
        super().__init__(self.message)

    def __str__(self):
        return self.message


class RuntimeException(PrismException):
    """
    Exception raised during command runtime.
    """

    def __init__(self, message):
        self.message = message
        super().__init__(self.message)

    def __str__(self):
        return self.message


class InvalidProfileException(PrismException):
    """
    Exception raised during command runtime.
    """

    def __init__(self, message):
        self.message = message
        super().__init__(self.message)

    def __str__(self):
        return self.message


class ProjectPyNotFoundException(PrismException):
    """
    Exception raised if the prism_project.py isn't found
    """

    def __init__(self, message):
        self.message = message
        super().__init__(self.message)

    def __str__(self):
        return self.message


class ModulesDirNotFoundException(PrismException):
    """
    Exception raised if modules dir isn't not found
    """

    def __init__(self, message):
        self.message = message
        super().__init__(self.message)

    def __str__(self):
        return self.message


class YamlNotFoundException(PrismException):
    """
    Exception raised if the profile.yml not found
    """

    def __init__(self, message):
        self.message = message
        super().__init__(self.message)

    def __str__(self):
        return self.message


class InvalidProjectPyException(PrismException):
    """
    Exception raised if parsing prism_project.py raises an error
    """
    def __init__(self, message):
        self.message = message
        super().__init__(self.message)

    def __str__(self):
        return self.message


class InvalidImportException(PrismException):
    """
    Exception raised if imports in prism_project.py are not specified correctly
    """
    def __init__(self, message):
        self.message = message
        super().__init__(self.message)

    def __str__(self):
        return self.message


class InvalidJinjaException(PrismException):
    """
    Exception raised when parsing jinja syntax in .yml files
    """
    def __init__(self, message, line_number):
        self.message = message
        self.line_number = line_number
        super().__init__(self.message)

    def __str__(self):
        return self.message


class CompileException(PrismException):
    """
    Exception raised during command runtime.
    """

    def __init__(self, message):
        self.message = message
        super().__init__(self.message)

    def __str__(self):
        return self.message


class DAGException(PrismException):
    """
    Exception raised during command runtime.
    """

    def __init__(self, message):
        self.message = message
        super().__init__(self.message)

    def __str__(self):
        return self.message


class ConsoleEventException(PrismException):
    """
    Exception raised during command runtime.
    """

    def __init__(self, message):
        self.message = message
        super().__init__(self.message)

    def __str__(self):
        return self.message


class ParserException(PrismException):
    """
    Exception raised during command runtime.
    """

    def __init__(self, message):
        self.message = message
        super().__init__(self.message)

    def __str__(self):
        return self.message


class InvalidTriggerException(PrismException):
    """
    Exception raised for invalid triggers.
    """

    def __init__(self, message):
        self.message = message
        super().__init__(self.message)

    def __str__(self):
        return self.message



from prism.task import PrismTask
import prism.target as PrismTarget

class Module13(prism.task.PrismTask):

    def run(self, tasks, hooks):
        return tasks.ref('module10.py') + "This is module 13. "


# EOF


###########
# Imports #
###########

# Prism infrastructure imports
import prism.task
import prism.target
import prism.decorators

# Prism project imports
import prism_project


######################
## Class definition ##
######################

class FilterCustomers(prism.task.PrismTask):

    ## Run    
    @prism.decorators.target(type=prism.target.PandasCsv, loc=prism_project.OUTPUT / 'bad_adapter.csv', index=False)
    def run(self, tasks, hooks):
        df = hooks.dbt_ref('dbt_prsdfofile', 'customers')
        df_new = df.iloc[:10]
        return df_new


# EOF


"""
Main

Table of Contents
- Imports
- Functions / utils
- Main
"""

###########
# Imports #
###########

import argparse
import prism.constants
from prism.cli import connect, init, run, compile, spark_submit, graph, trigger


##############
# Arg parser #
##############

# Key-value class
class KeyValue(argparse.Action):
    """
    Allow users to specify key-value pairs in their arguments. These will overwrite any
    variables in `prism_project.py`
    """

    def __call__(self,
        parser,
        namespace,
        values,
        option_string=None
    ):
        setattr(namespace, self.dest, dict())

        # Iterate through values, split each into k and value and assign to dictionary.
        for value in values:
            key, value = value.split('=')
            getattr(namespace, self.dest)[key] = value


def build_common_arguments_parser() -> argparse.ArgumentParser:
    """
    Build base subparser for common arguments across all commands.

    args:
        None
    returns:
        common_arguments_parser: instantiated ArgumentParser class with arguments for
        base subparser
    """
    common_arguments_parser = argparse.ArgumentParser(add_help=False)

    common_arguments_parser.add_argument(
        '--full-tb',
        required=False,
        action='store_true',
        help="""
        Display full traceback if errors arise at any stage of the pipeline
        """
    )

    common_arguments_parser.add_argument(
        '--log-level',
        required=False,
        type=str,
        default='info',
        help="""
        Set the log level; must be one of `info`, `warn`, `error`, or `critical`
        """
    )

    # Key-value variables
    common_arguments_parser.add_argument(
        '--vars',
        required=False,
        action=KeyValue,
        nargs='*',
        help="""
        Prism variables as key-value pairs `key=value`. These overwrite any variable
        definitions in `prism_project.py`. All values are read as strings.
        """
    )

    # Context JSON (as a string)
    common_arguments_parser.add_argument(
        '--context',
        required=False,
        type=str,
        default='{}',
        help="""
        Prism variables as JSON. Cannot co-exist with --vars. These overwrite any
        variable definitions in `prism_project.py`.
        """
    )
    return common_arguments_parser


def build_init_subparser(sub):
    """
    Build subparser for init command line argument.

    args:
        sub: special-action object (see argparse docs) to add subparsers to
        common_arguments_parser: parser with common arguments
    returns:
        None
    """
    init_sub = sub.add_parser(
        'init',
        help="""
        Initialize a pipeline
        """
    )

    # Add argument for project name
    init_sub.add_argument(
        '--project-name',
        type=str,
        help="""
        Project name
        """
    )

    # Add argument for whether to create minimal project
    init_sub.add_argument(
        '--minimal',
        required=False,
        action='store_true',
        help="""
        Create minimal project (just `prism_project.py` and `modules`)
        """
    )

    init_sub.add_argument(
        '--log-level',
        required=False,
        type=str,
        default='info',
        help="""
        Set the log level; must be one of `info`, `warn`, `error`, or `critical`
        """
    )

    # Set default class argument to InitTask()
    init_sub.set_defaults(cls=init.InitTask, which='init')


def build_connect_subparser(sub, common_arguments_parser):
    """
    Build subparser for connect command line argument.

    args:
        sub: special-action object (see argparse docs) to add subparsers to
        common_arguments_parser: parser with common arguments
    returns:
        None
    """
    connect_sub = sub.add_parser(
        'connect',
        parents=[common_arguments_parser],
        help="""
        Create a connection to a data warehouse or data processing system.
        At this time, prism supports connecting to Snowflake, PySpark, and Dbt.
        """
    )

    # Add argument for connection type
    valid_connections_str = ','.join([f'`{k}`' for k in prism.constants.VALID_ADAPTERS])
    connect_sub.add_argument(
        '--type',
        type=str,
        required=True,
        help=f"""
        Connection type. One of {valid_connections_str}
        """
    )

    # Set default class argument to RunTask()
    connect_sub.set_defaults(cls=connect.ConnectTask, which='connect')


def build_compile_subparser(sub, common_arguments_parser):
    """
    Build subparser for compile command line argument.

    args:
        sub: special-action object (see argparse docs) to add subparsers to
        common_arguments_parser: parser with common arguments
    returns:
        None
    """
    compile_sub = sub.add_parser(
        'compile',
        parents=[common_arguments_parser],
        help="""
        Parse pipeline inputs and outputs and create a DAG
        """
    )

    # Set default class argument to RunTask()
    compile_sub.set_defaults(cls=compile.CompileTask, modules=None, which='compile')


def build_run_subparser(sub, common_arguments_parser):
    """
    Build subparser for run command line argument.

    args:
        sub: special-action object (see argparse docs) to add subparsers to
        common_arguments_parser: parser with common arguments
    returns:
        None
    """
    run_sub = sub.add_parser(
        'run',
        parents=[common_arguments_parser],
        help="""
        Run a pipeline
        """
    )

    # Add argument for which script(s) to run
    run_sub.add_argument(
        '--modules',
        type=str,
        nargs='+',
        help="""
        Path to script(s) that you want to run; if not specified, all modules in
        pipeline are run
        """
    )

    # Add argument for whether to run all upstream modules
    run_sub.add_argument(
        '--all-upstream',
        required=False,
        action='store_true',
        help="""
        Run all modules upstream of explicit run set
        """
    )

    # Add argument for whether to run all upstream modules
    run_sub.add_argument(
        '--all-downstream',
        required=False,
        action='store_true',
        help="""
        Run all modules downstream of explicit run set
        """
    )

    # Set default class argument to RunTask()
    run_sub.set_defaults(cls=run.RunTask, which='run')


def build_spark_submit_subparser(sub, common_arguments_parser):
    """
    Build subparser for spark-submit command line argument.

    args:
        sub: special-action object (see argparse docs) to add subparsers to
        common_arguments_parser: parser with common arguments
    returns:
        None
    """
    spark_submit_sub = sub.add_parser(
        'spark-submit',
        parents=[common_arguments_parser],
        help="""
        Run a pipeline as a PySpark job
        """
    )

    # Add argument for which notebook to run
    spark_submit_sub.add_argument(
        '--modules',
        type=str,
        nargs='+',
        help="""
        Path to script(s) that you want to run; if not specified, all modules in
        pipeline are run
        """
    )

    # Add argument for whether to run all upstream modules
    spark_submit_sub.add_argument(
        '--all-upstream',
        required=False,
        action='store_true',
        help="""
        Run all modules upstream of explicit run set
        """
    )

    # Add argument for whether to run all downstream modules
    spark_submit_sub.add_argument(
        '--all-downstream',
        required=False,
        action='store_true',
        help="""
        Run all modules downstream of explicit run set
        """
    )

    # Set default class argument to RunTask()
    spark_submit_sub.set_defaults(
        cls=spark_submit.SparkSubmitTask, which='spark-submit'
    )


def build_graph_subparser(sub, common_arguments_parser):
    """
    Build subparser for graph command line argument.

    args:
        sub: special-action object (see argparse docs) to add subparsers to
        common_arguments_parser: parser with common arguments
    returns:
        None
    """
    graph_sub = sub.add_parser(
        'graph',
        parents=[common_arguments_parser],
        help="""
        Generate the DAG UI vizualizer
        """
    )

    # Add argument for the port to use
    graph_sub.add_argument(
        '--port',
        type=int,
        default=8080,
        help="""
        Port to use for HTTP request; default is 8080
        """
    )

    # Add argument for whether Prism should open a webbrowser with the dos
    graph_sub.add_argument(
        '--no-browser',
        action='store_true',
        help="""
        Overwrite default behavior and do not open a tab in the default web browser with
        the docs UI
        """
    )

    # Set default class argument to RunTask()
    graph_sub.set_defaults(cls=graph.GraphTask, all_downstream=True, which='graph')


def build_trigger_subparser(sub, common_arguments_parser):
    """
    Build subparser for trigger command line argument.

    args:
        sub: special-action object (see argparse docs) to add subparsers to
        common_arguments_parser: parser with common arguments
    returns:
        None
    """
    trigger_sub = sub.add_parser(
        'trigger',
        parents=[common_arguments_parser],
        help="""
        Create a trigger
        """
    )

    # Add argument for triggerion type
    valid_triggers_str = ','.join([f'`{k}`' for k in prism.constants.VALID_ADAPTERS])
    trigger_sub.add_argument(
        '--type',
        type=str,
        required=True,
        help=f"""
        Trigger type. One of {valid_triggers_str}
        """
    )

    # Set default class argument to RunTask()
    trigger_sub.set_defaults(cls=trigger.TriggerTask, which='trigger')


def build_full_arg_parser() -> argparse.ArgumentParser:
    """
    Build full argument parser
    """
    # Common arguments parser
    common_arguments_parser = build_common_arguments_parser()

    # Base parser
    base_parser = argparse.ArgumentParser()
    subparser = base_parser.add_subparsers()

    # Add subparsers
    build_init_subparser(subparser)
    build_run_subparser(subparser, common_arguments_parser)
    build_compile_subparser(subparser, common_arguments_parser)
    build_connect_subparser(subparser, common_arguments_parser)
    build_spark_submit_subparser(subparser, common_arguments_parser)
    build_graph_subparser(subparser, common_arguments_parser)
    build_trigger_subparser(subparser, common_arguments_parser)

    # Return base parser
    return base_parser


########
# Main #
########

def main(argv=None, bool_return: bool = False):
    """
    Parse argument, extract task, and execute
    """
    # Base parser
    base_parser = build_full_arg_parser()

    # Parse args
    parsed = base_parser.parse_args(argv)
    task = parsed.cls.task_from_args(parsed)
    result = task.run()

    # Return result
    if bool_return:
        return result
